/*
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License");  you may not use this file except in 
 * compliance with the License.  You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific language governing rights and limitations under the License.
 *
 * The Original Code is Protege-2000.
 *
 * The Initial Developer of the Original Code is Stanford University. Portions
 * created by Stanford University are Copyright (C) 2007.  All Rights Reserved.
 *
 * Protege was developed by Stanford Medical Informatics
 * (http://www.smi.stanford.edu) at the Stanford University School of Medicine
 * with support from the National Library of Medicine, the National Science
 * Foundation, and the Defense Advanced Research Projects Agency.  Current
 * information about Protege can be obtained at http://protege.stanford.edu.
 *
 */


package edu.stanford.smi.protegex.owl.swrl.ui.tab;

import java.awt.BorderLayout;
import java.awt.GridLayout;
import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.net.URI;
import java.util.Collection;
import java.util.Iterator;
import java.util.Set;
import java.util.logging.Level;

import javax.swing.JButton;
import javax.swing.JLabel;

import edu.stanford.smi.protege.model.Project;
import edu.stanford.smi.protege.util.Log;
import edu.stanford.smi.protege.widget.AbstractTabWidget;
import edu.stanford.smi.protegex.owl.jena.JenaOWLModel;
import edu.stanford.smi.protegex.owl.jena.parser.ProtegeOWLParser;
import edu.stanford.smi.protegex.owl.model.OWLModel;
import edu.stanford.smi.protegex.owl.model.OWLOntology;
import edu.stanford.smi.protegex.owl.model.RDFProperty;
import edu.stanford.smi.protegex.owl.model.factory.OWLJavaFactoryUpdater;
import edu.stanford.smi.protegex.owl.model.triplestore.TripleStore;
import edu.stanford.smi.protegex.owl.model.util.ImportHelper;
import edu.stanford.smi.protegex.owl.swrl.model.SWRLNames;
import edu.stanford.smi.protegex.owl.swrl.model.factory.SWRLJavaFactory;
import edu.stanford.smi.protegex.owl.swrl.ui.SWRLProjectPlugin;
import edu.stanford.smi.protegex.owl.swrl.ui.icons.SWRLIcons;
import edu.stanford.smi.protegex.owl.swrl.ui.table.SWRLTablePanel;
import edu.stanford.smi.protegex.owl.ui.ProtegeUI;

/**
 * A tab widget that holds the <a href="http://protege.cim3.net/cgi-bin/wiki.pl?SWRLEditorFAQ">SWRL Editor</a> and other plugins that work
 * with SWRL rules. This tab serves as the entry point to all of the GUI-based software components that work with SWRL in Protege-OWL. <p>
 *
 * Full documentation is available <a href="http://protege.cim3.net/cgi-bin/wiki.pl?SWRLTab">here</a>.
 */
public class SWRLTab extends AbstractTabWidget 
{ 
  private SWRLTablePanel panel = null;
  
  private void activateSWRL() 
  {
    JenaOWLModel owlModel = (JenaOWLModel) getKnowledgeBase();
    try {
      owlModel.getNamespaceManager().setPrefix(new URI(SWRLNames.SWRL_NAMESPACE), SWRLNames.SWRL_PREFIX);
      owlModel.getNamespaceManager().setPrefix(new URI(SWRLNames.SWRLB_NAMESPACE), SWRLNames.SWRLB_PREFIX);
      owlModel.getNamespaceManager().setPrefix(new URI(SWRLNames.SWRLA_NAMESPACE), SWRLNames.SWRLA_PREFIX);
      owlModel.getNamespaceManager().setPrefix(new URI(SWRLNames.SWRLX_NAMESPACE), SWRLNames.SWRLX_PREFIX);
      owlModel.getNamespaceManager().setPrefix(new URI(SWRLNames.SWRLTBOX_NAMESPACE), SWRLNames.SWRLTBOX_PREFIX);
      owlModel.getNamespaceManager().setPrefix(new URI(SWRLNames.SWRLABOX_NAMESPACE), SWRLNames.SWRLABOX_PREFIX);
      owlModel.getNamespaceManager().setPrefix(new URI(SWRLNames.SWRLQUERY_NAMESPACE), SWRLNames.SWRLQUERY_PREFIX);
      owlModel.getNamespaceManager().setPrefix(new URI(SWRLNames.SWRLTEMPORAL_NAMESPACE), SWRLNames.SWRLTEMPORAL_PREFIX);
  
      ImportHelper importHelper = new ImportHelper((JenaOWLModel)getKnowledgeBase());
      
      importHelper.addImport(new URI(SWRLNames.SWRL_IMPORT));     
      importHelper.addImport(new URI(SWRLNames.SWRLB_IMPORT));
      importHelper.addImport(new URI(SWRLNames.SWRLA_IMPORT));
      importHelper.addImport(new URI(SWRLNames.SWRLX_IMPORT));
      importHelper.addImport(new URI(SWRLNames.SWRLTBOX_IMPORT));
      importHelper.addImport(new URI(SWRLNames.SWRLABOX_IMPORT));
      importHelper.addImport(new URI(SWRLNames.SWRLQUERY_IMPORT));
      importHelper.addImport(new URI(SWRLNames.SWRLTEMPORAL_IMPORT));
      
      importHelper.importOntologies();

      // Make ":TO" and ":FROM" visible for dynamic expansion.
      RDFProperty to = owlModel.getOWLObjectProperty(edu.stanford.smi.protege.model.Model.Slot.TO);
      to.setVisible(true);
      RDFProperty from = owlModel.getOWLObjectProperty(edu.stanford.smi.protege.model.Model.Slot.FROM);
      from.setVisible(true);
    } catch (Exception ex) {
      ProtegeUI.getModalDialogFactory().showErrorMessageDialog(owlModel,
                                                               "Could not activate SWRLTab: " + ex +
                                                               "\n. Your project might be in an inconsistent state now.");
      Log.getLogger().log(Level.SEVERE, "Exception caught", ex);
    }
  }
  
  public void initialize() 
  {
    setLabel("SWRL Rules");
    setIcon(SWRLIcons.getImpsIcon());
    if (!(getKnowledgeBase() instanceof OWLModel)) {
      add(BorderLayout.CENTER, new JLabel("This tab can only be used with OWL projects."));
      return;
    } // if
    
    OWLModel owlModel = (OWLModel) getKnowledgeBase();
    
    if (!SWRLProjectPlugin.isSWRLImported(owlModel)) {
      if (isSWRLImported(owlModel)) {
        activateSWRLFactoryIfNecessary(owlModel);
      } else {
        setLayout(new FlowLayout());
        add(new JLabel("Your ontology needs to import the SWRL ontology (" + SWRLNames.SWRL_NAMESPACE + ")."));
        JButton activateButton = new JButton("Activate SWRL...");
        activateButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
              activateSWRL();
            }
          });
        add(activateButton);
        if (!(owlModel instanceof JenaOWLModel)) {
          activateButton.setEnabled(false);
        }        		
      } // if
    } // if
        
    // Have to check again because the activateSWRLFactoryIfNecessary may have failed
    if (SWRLProjectPlugin.isSWRLImported(owlModel)) {
      panel = new SWRLTablePanel(owlModel, null, this);
      add(panel);
    }        
    
    SWRLProjectPlugin.adjustGUI(getProject());
  } // initialize 

  public void reconfigure()
  {
    if (panel != null) {
      remove(panel);
      setLayout(new BorderLayout());
      add(panel);
    } // if
  } // reconfigure

  private void activateSWRLFactoryIfNecessary(OWLModel owlModel) 
  {	
    SWRLJavaFactory factory = new SWRLJavaFactory(owlModel);
    owlModel.setOWLJavaFactory(factory);
    if(owlModel instanceof JenaOWLModel) {
      OWLJavaFactoryUpdater.run((JenaOWLModel) owlModel);
    }	
  }

  private boolean isSWRLImported(OWLModel owlModel) 
  {
    boolean swrlFound = false;
    boolean swrlbFound = false;
    
    Iterator iter = owlModel.getOWLOntologies().iterator();
    
    try {
      while (iter.hasNext() && !(swrlbFound && swrlFound)) {
        OWLOntology ont = (OWLOntology) iter.next();
        if (ont.getNamespace().equals(SWRLNames.SWRL_NAMESPACE)) swrlFound = true;
        if (ont.getNamespace().equals(SWRLNames.SWRLB_NAMESPACE)) swrlbFound = true;				
      }
    } catch (Exception e) {
      // TODO: handle exception
    }
    return swrlFound && swrlbFound;
  } // isSWRLImported
  
  public static boolean isSuitable(Project p, Collection errors) 
  {
    if (p.getKnowledgeBase() instanceof OWLModel) {
      return true;
    } else {
      errors.add("This tab can only be used with OWL projects.");
      return false;
    }
  }
} // SWRLTab
